#!/bin/bash

# Package Sidef as a binary executable
# Requires: App::Packer::PAR

/usr/bin/site_perl/pp --compile --unicode --execute --lib=../lib -M Sidef::Types::Bool::Bool -M Sidef -M Sidef::Parser -M Sidef::Optimizer -M Sidef::Object::Object -M Sidef::Types::Number::Number -M Sidef::Deparse::Perl -M Sidef::Types::Block::Block -M Memoize -M Sidef::Types::Number::Complex -M Sidef::Types::Array::Array -M Sidef::Types::Hash::Hash -M Sidef::Perl::Perl -M Sidef::Sys::Sys -M Sidef::Math::Math -M Sidef::Types::String::String -M Sidef::Types::Range::RangeNumber -M Sidef::Types::Range::RangeString -M Sidef::Types::Glob::File -M Sidef::Types::Glob::Dir -M Sidef::Types::Glob::FileHandle -M Sidef::Types::Glob::DirHandle -M Sidef -M Sidef::Types::Regex::Regex -M Sidef::Types::Regex::Match -M Sidef::Module::Func -M Sidef::Module::OO -M Sidef::Sys::Sig -M Sidef::Time::Time -M Sidef::Time::Localtime -M Sidef::Time::Gmtime -M Sidef::Object::LazyMethod -M Sidef::Variable::NamedParam -M Sidef::Variable::GetOpt -M Sidef::Object::Convert -M Sidef::Types::Null::Null -M Sidef::Types::Glob::Backtick -M Sidef::Types::Glob::Pipe -M Sidef::Types::Glob::Socket -M Sidef::Types::Glob::SocketHandle -M Sidef::Types::Glob::Stat -M Sidef::Types::Block::Fork -M Sidef::Types::Block::Try -M Sidef::Types::Array::Pair -M Sidef::Deparse::Sidef -M experimental -M base -M File::Spec -M File::Temp -M File::Basename -M File::Compare -M Cwd -M Time::HiRes -M Scalar::Util -M Socket -M Encode -M Fcntl -M POSIX -M File::Find -M File::Copy -M File::Path -M utf8 -M List::Util -M Math::MPFR -M Math::GMPz -M Math::GMPq -M Math::MPC -M Sidef::Object::Lazy -M Sidef::Object::Enumerator -M charnames -M bytes -M Data::Dump -M Data::Dump::Filtered -M Sidef::Types::Range::Range -M Math::Prime::Util::GMP -M Getopt::Long -M Digest::MD5 -M Digest::SHA -M Algorithm::Combinatorics -M Algorithm::Loops -M MIME::Base64 -M File::HomeDir -M IO -M Encode::Encoding -M Encode::Unicode -M Exporter -M Exporter::Heavy -M Fcntl -M Cwd -M Config -M Encode::Alias -M Encode::MIME::Name -M File::Glob -M File::Spec::Unix -M File::Spec::Win32 -M IO::Handle -M DynaLoader -M re -M POSIX -M SelectSaver -M SelfLoader -M Term::Cap -M Term::ReadKey -M Term::ReadLine -M Tie::Hash -M Tie::Hash::NamedCapture -M utf8 -M XSLoader -M Symbol -M Storable -M Time::HiRes -M overload -M overloading -M parent -M Sidef::Types::Set::Set -M Sidef::Types::Set::Bag -M Sidef::Types::Array::Matrix -M Sidef::Types::Array::Vector -M Data::Dump -M Carp -M bytes -M base -M constant -M charnames -M _charnames -M integer -M unicore::Name -M vars -M warnings -M warnings::register -M subs -M List::Util -M DB_File -M Math::Prime::Util -o sidef.out -z 9 -c ../bin/sidef
